﻿using System;
using System.Globalization;
using System.Threading;
using FireBus.Contracts;
using NUnit.Framework;

using FireBus.UnitTest.Contracts;
using FireBus.UnitTest.Fixtures.Hosts;


namespace FireBus.UnitTest.Fixtures
{
	[TestFixture]
	public class HostTestFixture
	{
		[Test]
		public void Request_ThroughRequestSender_RequestShouldBeSentThroughAllRequestListeners()
		{
			using (var host = new DefaultTestHost())
			{
				host.AddUpperCaseConverter("upper");
				host.AddStringConverter("converter");
				var asyncResult = host.BeginLoadAllExtensions(null, null);
				host.EndLoadAllExtensions(asyncResult);

				asyncResult = host.BeginRequest<StringConversionRequest, String>(
					"converter", new StringConversionRequest("test", "upper"), null, null);
				var output = host.EndRequest<String>(asyncResult);

				Assert.IsNotNull(output);
				Assert.AreEqual("TEST", output);
			}
		}

		[Test]
		public void NotifyAll__ShouldNotifyAllEventListeners()
		{
			using (var host = new DefaultTestHost())
			{
				//Adds several log extensions
				const Int32 numberOfLogExtensions = 5;
				for (var i = 0; i < numberOfLogExtensions; i++)
				{
					var name = String.Format(
						CultureInfo.CurrentCulture,
						"log_{0}",
						i);
					host.AddLog(name);
				}
				//Loads extensions and initializes them with log storage
				var logStorage = new LogStorage();
				foreach (var extensionToken in host.GetExtensions())
				{
					host.BeginLoadExtension(
						extensionToken.Name, null, null).AsyncWaitHandle.WaitOne();

					var startup = new StartupEvent(extensionToken.Name, logStorage);
					host.BeginNotify(extensionToken.Name, startup, null, null).AsyncWaitHandle.WaitOne();
				}

				//Sends message to all log extensions
				var message =
					new LogMessageEvent(
						"test",
						"testfixture",
						LogMessageType.Notification,
						DateTime.Now);
				var asyncResult = host.BeginNotifyAll(message, null, null);
				host.EndNotifyAll(asyncResult);

				//Checks if message was logged by all log extensions.
				foreach (var extensionToken in host.GetExtensions())
				{
					Assert.True(
						logStorage.Contains(
							extensionToken.Name,
							message),
						String.Format(
							CultureInfo.CurrentCulture,
							"Message was not logged by extension {0}",
							extensionToken.Name));
				}
			}
		}

		[Test]
		public void NotifyAll_ForAnUnknownToHostEvent_ShouldNotifyAllEventListeners()
		{
			using (var host = new DefaultTestHost())
			{
				#region Token Initialization

				var notifier =
					new ExtensionToken(
						new ExtensionName("notifier"),
						new ExtensionPath(
							@"..\..\..\FireBus.UnitTest.Extensions\bin\Debug\"),
						new TypeName(
							"FireBus.UnitTest.Extensions.EventNotifierExtension, FireBus.UnitTest.Extensions, Version=0.1.0.0"));
				host.AddExtension(notifier);
				var listener =
					new ExtensionToken(
						new ExtensionName("listener"),
						new ExtensionPath(
							@"..\..\..\FireBus.UnitTest.Extensions\bin\Debug\"),
						new TypeName(
							"FireBus.UnitTest.Extensions.EventListenerExtension, FireBus.UnitTest.Extensions, Version=0.1.0.0"));
				host.AddExtension(listener);

				#endregion Token Initialization

				host.BeginLoadExtension(notifier.Name, null, null).AsyncWaitHandle.WaitOne();
				host.BeginLoadExtension(listener.Name, null, null).AsyncWaitHandle.WaitOne();

				var hostAccessor = new HostAccessor(host);
				var logStorage = new LogStorage();

				host.BeginNotify(
					listener.Name,
					new StartupWithHostEvent(
						hostAccessor, listener.Name, logStorage),
					null,
					null).AsyncWaitHandle.WaitOne();

				Assert.IsTrue(
					logStorage.Count == 0,
					"Log storage should be empty before EventSource was started");
				host.BeginNotify(
					notifier.Name,
					new StartupWithHostEvent(
						hostAccessor, notifier.Name, logStorage),
					null,
					null).AsyncWaitHandle.WaitOne();

				Thread.Sleep(1000);

				Assert.IsTrue(
					logStorage.Count == 1,
					"Log storage should contain one event after EventNotifier was started");
			}
		}
	}
}
